@@ -3,35 +3,52 @@
 '''Data dispatching for DMatrix.'''
 import ctypes
 import json
-import warnings
 import os
-from typing import Any, Tuple, Callable, Optional, List, Union, Iterator
+import warnings
+from typing import Any, Callable, Iterator, List, Optional, Sequence, Tuple, Union, cast
 
 import numpy as np
 
-from .core import c_array, _LIB, _check_call, c_str
-from .core import _cuda_array_interface
-from .core import DataIter, _ProxyDMatrix, DMatrix, FeatureNames
-from .compat import lazy_isinstance, DataFrame
+from ._typing import (
+    CupyT,
+    DataType,
+    FeatureNames,
+    FeatureTypes,
+    FloatCompatible,
+    NumpyDType,
+    PandasDType,
+    c_bst_ulong,
+)
+from .compat import DataFrame, lazy_isinstance
+from .core import (
+    _LIB,
+    DataIter,
+    DMatrix,
+    _check_call,
+    _cuda_array_interface,
+    _ProxyDMatrix,
+    c_array,
+    c_str,
+    from_pystr_to_cstr,
+)
 
-c_bst_ulong = ctypes.c_uint64   # pylint: disable=invalid-name
+DispatchedDataBackendReturnType = Tuple[
+    ctypes.c_void_p, Optional[FeatureNames], Optional[FeatureTypes]]
 
 CAT_T = "c"
 
 # meta info that can be a matrix instead of vector.
-# For now it's base_margin for multi-class, but it can be extended to label once we have
-# multi-output.
 _matrix_meta = {"base_margin", "label"}
 
 
-def _warn_unused_missing(data, missing):
+def _warn_unused_missing(data: DataType, missing: Optional[FloatCompatible]) -> None:
     if (missing is not None) and (not np.isnan(missing)):
         warnings.warn(
             '`missing` is not used for current input data type:' +
             str(type(data)), UserWarning)
 
 
-def _check_complex(data):
+def _check_complex(data: DataType) -> None:
     '''Test whether data is complex using `dtype` attribute.'''
     complex_dtypes = (np.complex128, np.complex64,
                       np.cfloat, np.cdouble, np.clongdouble)
@@ -39,16 +56,15 @@ def _check_complex(data):
         raise ValueError('Complex data not supported')
 
 
-def _check_data_shape(data: Any) -> None:
+def _check_data_shape(data: DataType) -> None:
     if hasattr(data, "shape") and len(data.shape) != 2:
         raise ValueError("Please reshape the input data into 2-dimensional matrix.")
 
 
-def _is_scipy_csr(data):
+def _is_scipy_csr(data: DataType) -> bool:
     try:
-        import scipy
+        import scipy.sparse
     except ImportError:
-        scipy = None
         return False
     return isinstance(data, scipy.sparse.csr_matrix)
 
@@ -65,12 +81,12 @@ def _array_interface(data: np.ndarray) -> bytes:
 
 
 def _from_scipy_csr(
-    data,
-    missing,
-    nthread,
-    feature_names: FeatureNames,
-    feature_types: Optional[List[str]],
-):
+    data: DataType,
+    missing: FloatCompatible,
+    nthread: int,
+    feature_names: Optional[FeatureNames],
+    feature_types: Optional[FeatureTypes],
+) -> DispatchedDataBackendReturnType:
     """Initialize data from a CSR matrix."""
     if len(data.indices) != len(data.data):
         raise ValueError(
@@ -87,7 +103,7 @@ def _from_scipy_csr(
             _array_interface(data.indptr),
             _array_interface(data.indices),
             _array_interface(data.data),
-            ctypes.c_size_t(data.shape[1]),
+            c_bst_ulong(data.shape[1]),
             config,
             ctypes.byref(handle),
         )
@@ -95,21 +111,20 @@ def _from_scipy_csr(
     return handle, feature_names, feature_types
 
 
-def _is_scipy_csc(data):
+def _is_scipy_csc(data: DataType) -> bool:
     try:
-        import scipy
+        import scipy.sparse
     except ImportError:
-        scipy = None
         return False
     return isinstance(data, scipy.sparse.csc_matrix)
 
 
 def _from_scipy_csc(
-    data,
-    missing,
-    feature_names: FeatureNames,
-    feature_types: Optional[List[str]],
-):
+    data: DataType,
+    missing: Optional[FloatCompatible],
+    feature_names: Optional[FeatureNames],
+    feature_types: Optional[FeatureTypes],
+) -> DispatchedDataBackendReturnType:
     if len(data.indices) != len(data.data):
         raise ValueError(f"length mismatch: {len(data.indices)} vs {len(data.data)}")
     _warn_unused_missing(data, missing)
@@ -125,27 +140,29 @@ def _from_scipy_csc(
     return handle, feature_names, feature_types
 
 
-def _is_scipy_coo(data):
+def _is_scipy_coo(data: DataType) -> bool:
     try:
-        import scipy
+        import scipy.sparse
     except ImportError:
-        scipy = None
         return False
     return isinstance(data, scipy.sparse.coo_matrix)
 
 
-def _is_numpy_array(data):
+def _is_numpy_array(data: DataType) -> bool:
     return isinstance(data, (np.ndarray, np.matrix))
 
 
-def _ensure_np_dtype(data, dtype) -> Tuple[np.ndarray, np.dtype]:
+def _ensure_np_dtype(
+    data: DataType,
+    dtype: Optional[NumpyDType]
+) -> Tuple[np.ndarray, Optional[NumpyDType]]:
     if data.dtype.hasobject or data.dtype in [np.float16, np.bool_]:
         data = data.astype(np.float32, copy=False)
         dtype = np.float32
     return data, dtype
 
 
-def _maybe_np_slice(data: np.ndarray, dtype) -> np.ndarray:
+def _maybe_np_slice(data: DataType, dtype: Optional[NumpyDType]) -> np.ndarray:
     '''Handle numpy slice.  This can be removed if we use __array_interface__.
     '''
     try:
@@ -160,12 +177,12 @@ def _maybe_np_slice(data: np.ndarray, dtype) -> np.ndarray:
 
 
 def _from_numpy_array(
-    data,
-    missing,
-    nthread,
-    feature_names: FeatureNames,
-    feature_types: Optional[List[str]],
-):
+    data: DataType,
+    missing: FloatCompatible,
+    nthread: int,
+    feature_names: Optional[FeatureNames],
+    feature_types: Optional[FeatureTypes],
+) -> DispatchedDataBackendReturnType:
     """Initialize data from a 2-D numpy matrix.
 
     """
@@ -190,7 +207,7 @@ def _from_numpy_array(
     return handle, feature_names, feature_types
 
 
-def _is_pandas_df(data):
+def _is_pandas_df(data: DataType) -> bool:
     try:
         import pandas as pd
     except ImportError:
@@ -198,7 +215,7 @@ def _is_pandas_df(data):
     return isinstance(data, pd.DataFrame)
 
 
-def _is_modin_df(data):
+def _is_modin_df(data: DataType) -> bool:
     try:
         import modin.pandas as pd
     except ImportError:
@@ -223,16 +240,24 @@ _pandas_dtype_mapper = {
     "Int16": "int",
     "Int32": "int",
     "Int64": "int",
+    "Float32": "float",
+    "Float64": "float",
     "boolean": "i",
 }
 
 
-def _invalid_dataframe_dtype(data: Any) -> None:
+_ENABLE_CAT_ERR = (
+    "When categorical type is supplied, The experimental DMatrix parameter"
+    "`enable_categorical` must be set to `True`."
+)
+
+
+def _invalid_dataframe_dtype(data: DataType) -> None:
     # pandas series has `dtypes` but it's just a single object
     # cudf series doesn't have `dtypes`.
     if hasattr(data, "dtypes") and hasattr(data.dtypes, "__iter__"):
         bad_fields = [
-            str(data.columns[i])
+            f"{data.columns[i]}: {dtype}"
             for i, dtype in enumerate(data.dtypes)
             if dtype.name not in _pandas_dtype_mapper
         ]
@@ -240,47 +265,24 @@ def _invalid_dataframe_dtype(data: Any) -> None:
     else:
         err = ""
 
-    msg = """DataFrame.dtypes for data must be int, float, bool or category.  When
-categorical type is supplied, DMatrix parameter `enable_categorical` must
-be set to `True`.""" + err
+    type_err = "DataFrame.dtypes for data must be int, float, bool or category."
+    msg = f"""{type_err} {_ENABLE_CAT_ERR} {err}"""
     raise ValueError(msg)
 
 
-# pylint: disable=too-many-locals
-def _transform_pandas_df(
+def _pandas_feature_info(
     data: DataFrame,
+    meta: Optional[str],
+    feature_names: Optional[FeatureNames],
+    feature_types: Optional[FeatureTypes],
     enable_categorical: bool,
-    feature_names: FeatureNames = None,
-    feature_types: Optional[List[str]] = None,
-    meta: Optional[str] = None,
-    meta_type: Optional[str] = None,
-) -> Tuple[np.ndarray, FeatureNames, Optional[List[str]]]:
+) -> Tuple[Optional[FeatureNames], Optional[FeatureTypes]]:
     import pandas as pd
     from pandas.api.types import (
         is_sparse,
         is_categorical_dtype,
-        is_integer_dtype,
-        is_bool_dtype,
     )
 
-    nullable_alias = {"Int16", "Int32", "Int64"}
-
-    # dtype: pd.core.arrays.numeric.NumericDtype
-    def is_nullable_dtype(dtype: Any) -> bool:
-        is_int = is_integer_dtype(dtype) and dtype.name in nullable_alias
-        # np.bool has alias `bool`, while pd.BooleanDtype has `boolean`.
-        is_bool = is_bool_dtype(dtype) and dtype.name == "boolean"
-        return is_int or is_bool
-
-    if not all(
-        dtype.name in _pandas_dtype_mapper
-        or is_sparse(dtype)
-        or is_nullable_dtype(dtype)
-        or (is_categorical_dtype(dtype) and enable_categorical)
-        for dtype in data.dtypes
-    ):
-        _invalid_dataframe_dtype(data)
-
     # handle feature names
     if feature_names is None and meta is None:
         if isinstance(data.columns, pd.MultiIndex):
@@ -293,61 +295,119 @@ def _transform_pandas_df(
     # handle feature types
     if feature_types is None and meta is None:
         feature_types = []
-        for i, dtype in enumerate(data.dtypes):
+        for dtype in data.dtypes:
             if is_sparse(dtype):
                 feature_types.append(_pandas_dtype_mapper[dtype.subtype.name])
             elif is_categorical_dtype(dtype) and enable_categorical:
                 feature_types.append(CAT_T)
             else:
                 feature_types.append(_pandas_dtype_mapper[dtype.name])
+    return feature_names, feature_types
 
-    # handle category codes.
-    transformed = pd.DataFrame()
-    # Avoid transformation due to: PerformanceWarning: DataFrame is highly fragmented
-    if (
-        enable_categorical and any(is_categorical_dtype(dtype) for dtype in data.dtypes)
-    ) or any(is_nullable_dtype(dtype) for dtype in data.dtypes):
-        for i, dtype in enumerate(data.dtypes):
-            if is_categorical_dtype(dtype):
-                # pandas uses -1 as default missing value for categorical data
-                transformed[data.columns[i]] = (
-                    data[data.columns[i]]
-                    .cat.codes.astype(np.float32)
-                    .replace(-1.0, np.NaN)
-                )
-            elif is_nullable_dtype(dtype):
-                # Converts integer <NA> to float NaN
-                transformed[data.columns[i]] = data[data.columns[i]].astype(np.float32)
-            else:
-                transformed[data.columns[i]] = data[data.columns[i]]
+
+def is_nullable_dtype(dtype: PandasDType) -> bool:
+    """Wether dtype is a pandas nullable type."""
+    from pandas.api.types import (
+        is_integer_dtype,
+        is_bool_dtype,
+        is_float_dtype,
+        is_categorical_dtype,
+    )
+
+    # dtype: pd.core.arrays.numeric.NumericDtype
+    nullable_alias = {"Int16", "Int32", "Int64", "Float32", "Float64", "category"}
+    is_int = is_integer_dtype(dtype) and dtype.name in nullable_alias
+    # np.bool has alias `bool`, while pd.BooleanDtype has `bzoolean`.
+    is_bool = is_bool_dtype(dtype) and dtype.name == "boolean"
+    is_float = is_float_dtype(dtype) and dtype.name in nullable_alias
+    return is_int or is_bool or is_float or is_categorical_dtype(dtype)
+
+
+def _pandas_cat_null(data: DataFrame) -> DataFrame:
+    from pandas.api.types import is_categorical_dtype
+    # handle category codes and nullable.
+    cat_columns = [
+        col
+        for col, dtype in zip(data.columns, data.dtypes)
+        if is_categorical_dtype(dtype)
+    ]
+    nul_columns = [
+        col for col, dtype in zip(data.columns, data.dtypes) if is_nullable_dtype(dtype)
+    ]
+    if cat_columns or nul_columns:
+        # Avoid transformation due to: PerformanceWarning: DataFrame is highly
+        # fragmented
+        transformed = data.copy()
     else:
         transformed = data
 
+    if cat_columns:
+        # DF doesn't have the cat attribute, so we use apply here
+        transformed[cat_columns] = (
+            transformed[cat_columns]
+            .apply(lambda x: x.cat.codes)
+            .astype(np.float32)
+            .replace(-1.0, np.NaN)
+        )
+    if nul_columns:
+        transformed[nul_columns] = transformed[nul_columns].astype(np.float32)
+
+    return transformed
+
+
+def _transform_pandas_df(
+    data: DataFrame,
+    enable_categorical: bool,
+    feature_names: Optional[FeatureNames] = None,
+    feature_types: Optional[FeatureTypes] = None,
+    meta: Optional[str] = None,
+    meta_type: Optional[NumpyDType] = None,
+) -> Tuple[np.ndarray, Optional[FeatureNames], Optional[FeatureTypes]]:
+    from pandas.api.types import (
+        is_sparse,
+        is_categorical_dtype,
+    )
+
+    if not all(
+        dtype.name in _pandas_dtype_mapper
+        or is_sparse(dtype)
+        or (is_nullable_dtype(dtype) and not is_categorical_dtype(dtype))
+        or (is_categorical_dtype(dtype) and enable_categorical)
+        for dtype in data.dtypes
+    ):
+        _invalid_dataframe_dtype(data)
+
+    feature_names, feature_types = _pandas_feature_info(
+        data, meta, feature_names, feature_types, enable_categorical
+    )
+
+    transformed = _pandas_cat_null(data)
+
     if meta and len(data.columns) > 1 and meta not in _matrix_meta:
         raise ValueError(f"DataFrame for {meta} cannot have multiple columns")
 
     dtype = meta_type if meta_type else np.float32
-    arr = transformed.values
+    arr: np.ndarray = transformed.values
     if meta_type:
-        arr = arr.astype(meta_type)
+        arr = arr.astype(dtype)
     return arr, feature_names, feature_types
 
 
 def _from_pandas_df(
     data: DataFrame,
     enable_categorical: bool,
-    missing: float,
+    missing: FloatCompatible,
     nthread: int,
-    feature_names: FeatureNames,
-    feature_types: Optional[List[str]],
-) -> Tuple[ctypes.c_void_p, FeatureNames, Optional[List[str]]]:
+    feature_names: Optional[FeatureNames],
+    feature_types: Optional[FeatureTypes],
+) -> DispatchedDataBackendReturnType:
     data, feature_names, feature_types = _transform_pandas_df(
         data, enable_categorical, feature_names, feature_types
     )
     return _from_numpy_array(data, missing, nthread, feature_names, feature_types)
 
 
-def _is_pandas_series(data):
+def _is_pandas_series(data: DataType) -> bool:
     try:
         import pandas as pd
     except ImportError:
@@ -356,18 +416,21 @@ def _is_pandas_series(data):
 
 
 def _meta_from_pandas_series(
-    data, name: str, dtype: Optional[str], handle: ctypes.c_void_p
+    data: DataType,
+    name: str,
+    dtype: Optional[NumpyDType],
+    handle: ctypes.c_void_p
 ) -> None:
     """Help transform pandas series for meta data like labels"""
     data = data.values.astype('float')
     from pandas.api.types import is_sparse
     if is_sparse(data):
-        data = data.to_dense()
+        data = data.to_dense()  # type: ignore
     assert len(data.shape) == 1 or data.shape[1] == 0 or data.shape[1] == 1
     _meta_from_numpy(data, name, dtype, handle)
 
 
-def _is_modin_series(data):
+def _is_modin_series(data: DataType) -> bool:
     try:
         import modin.pandas as pd
     except ImportError:
@@ -376,13 +439,13 @@ def _is_modin_series(data):
 
 
 def _from_pandas_series(
-    data,
-    missing: float,
+    data: DataType,
+    missing: FloatCompatible,
     nthread: int,
     enable_categorical: bool,
-    feature_names: FeatureNames,
-    feature_types: Optional[List[str]],
-):
+    feature_names: Optional[FeatureNames],
+    feature_types: Optional[FeatureTypes],
+) -> DispatchedDataBackendReturnType:
     from pandas.api.types import is_categorical_dtype
 
     if (data.dtype.name not in _pandas_dtype_mapper) and not (
@@ -400,7 +463,7 @@ def _from_pandas_series(
     )
 
 
-def _is_dt_df(data):
+def _is_dt_df(data: DataType) -> bool:
     return lazy_isinstance(data, 'datatable', 'Frame') or \
         lazy_isinstance(data, 'datatable', 'DataTable')
 
@@ -410,12 +473,12 @@ _dt_type_mapper2 = {'bool': 'i', 'int': 'int', 'real': 'float'}
 
 
 def _transform_dt_df(
-    data,
-    feature_names: FeatureNames,
-    feature_types: Optional[List[str]],
-    meta=None,
-    meta_type=None,
-):
+    data: DataType,
+    feature_names: Optional[FeatureNames],
+    feature_types: Optional[FeatureTypes],
+    meta: Optional[str] = None,
+    meta_type: Optional[NumpyDType] = None,
+) -> Tuple[np.ndarray, Optional[FeatureNames], Optional[FeatureTypes]]:
     """Validate feature names and types if data table"""
     if meta and data.shape[1] > 1:
         raise ValueError('DataTable for meta info cannot have multiple columns')
@@ -449,13 +512,13 @@ def _transform_dt_df(
 
 
 def _from_dt_df(
-    data,
-    missing,
-    nthread,
-    feature_names: FeatureNames,
-    feature_types: Optional[List[str]],
+    data: DataType,
+    missing: Optional[FloatCompatible],
+    nthread: int,
+    feature_names: Optional[FeatureNames],
+    feature_types: Optional[FeatureTypes],
     enable_categorical: bool,
-) -> Tuple[ctypes.c_void_p, FeatureNames, Optional[List[str]]]:
+) -> DispatchedDataBackendReturnType:
     if enable_categorical:
         raise ValueError("categorical data in datatable is not supported yet.")
     data, feature_names, feature_types = _transform_dt_df(
@@ -492,7 +555,7 @@ def _from_dt_df(
     return handle, feature_names, feature_types
 
 
-def _is_arrow(data) -> bool:
+def _is_arrow(data: DataType) -> bool:
     try:
         import pyarrow as pa
         from pyarrow import dataset as arrow_dataset
@@ -505,6 +568,7 @@ def record_batch_data_iter(data_iter: Iterator) -> Callable:
     """Data iterator used to ingest Arrow columnar record batches. We are not using
     class DataIter because it is only intended for building Device DMatrix and external
     memory DMatrix.
+
     """
     from pyarrow.cffi import ffi
 
@@ -537,13 +601,13 @@ def record_batch_data_iter(data_iter: Iterator) -> Callable:
 
 
 def _from_arrow(
-    data,
-    missing: float,
+    data: DataType,
+    missing: FloatCompatible,
     nthread: int,
-    feature_names: Optional[List[str]],
-    feature_types: Optional[List[str]],
+    feature_names: Optional[FeatureNames],
+    feature_types: Optional[FeatureTypes],
     enable_categorical: bool,
-) -> Tuple[ctypes.c_void_p, Optional[List[str]], Optional[List[str]]]:
+) -> DispatchedDataBackendReturnType:
     import pyarrow as pa
 
     if not all(
@@ -555,12 +619,14 @@ def _from_arrow(
     if enable_categorical:
         raise ValueError("categorical data in arrow is not supported yet.")
 
-    rb_iter = iter(data.to_batches())
+    batches = data.to_batches()
+    rb_iter = iter(batches)
     it = record_batch_data_iter(rb_iter)
     next_callback = ctypes.CFUNCTYPE(ctypes.c_int, ctypes.c_void_p)(it)
     handle = ctypes.c_void_p()
-
-    config = bytes(json.dumps({"missing": missing, "nthread": nthread}), "utf-8")
+    config = from_pystr_to_cstr(
+        json.dumps({"missing": missing, "nthread": nthread, "nbatch": len(batches)})
+    )
     _check_call(
         _LIB.XGDMatrixCreateFromArrowCallback(
             next_callback,
@@ -571,15 +637,15 @@ def _from_arrow(
     return handle, feature_names, feature_types
 
 
-def _is_cudf_df(data) -> bool:
+def _is_cudf_df(data: DataType) -> bool:
     return lazy_isinstance(data, "cudf.core.dataframe", "DataFrame")
 
 
-def _cudf_array_interfaces(data, cat_codes: list) -> bytes:
-    """Extract CuDF __cuda_array_interface__.  This is special as it returns a new list of
-    data and a list of array interfaces.  The data is list of categorical codes that
-    caller can safely ignore, but have to keep their reference alive until usage of array
-    interface is finished.
+def _cudf_array_interfaces(data: DataType, cat_codes: list) -> bytes:
+    """Extract CuDF __cuda_array_interface__.  This is special as it returns a new list
+    of data and a list of array interfaces.  The data is list of categorical codes that
+    caller can safely ignore, but have to keep their reference alive until usage of
+    array interface is finished.
 
     """
     try:
@@ -588,14 +654,18 @@ def _cudf_array_interfaces(data, cat_codes: list) -> bytes:
         from cudf.utils.dtypes import is_categorical_dtype
 
     interfaces = []
+
+    def append(interface: dict) -> None:
+        if "mask" in interface:
+            interface["mask"] = interface["mask"].__cuda_array_interface__
+        interfaces.append(interface)
+
     if _is_cudf_ser(data):
         if is_categorical_dtype(data.dtype):
             interface = cat_codes[0].__cuda_array_interface__
         else:
             interface = data.__cuda_array_interface__
-        if "mask" in interface:
-            interface["mask"] = interface["mask"].__cuda_array_interface__
-        interfaces.append(interface)
+        append(interface)
     else:
         for i, col in enumerate(data):
             if is_categorical_dtype(data[col].dtype):
@@ -603,19 +673,17 @@ def _cudf_array_interfaces(data, cat_codes: list) -> bytes:
                 interface = codes.__cuda_array_interface__
             else:
                 interface = data[col].__cuda_array_interface__
-            if "mask" in interface:
-                interface["mask"] = interface["mask"].__cuda_array_interface__
-            interfaces.append(interface)
-    interfaces_str = bytes(json.dumps(interfaces, indent=2), "utf-8")
+            append(interface)
+    interfaces_str = from_pystr_to_cstr(json.dumps(interfaces))
     return interfaces_str
 
 
 def _transform_cudf_df(
-    data,
-    feature_names: FeatureNames,
-    feature_types: Optional[List[str]],
+    data: DataType,
+    feature_names: Optional[FeatureNames],
+    feature_types: Optional[FeatureTypes],
     enable_categorical: bool,
-):
+) -> Tuple[ctypes.c_void_p, list, Optional[FeatureNames], Optional[FeatureTypes]]:
     try:
         from cudf.api.types import is_categorical_dtype
     except ImportError:
@@ -667,21 +735,26 @@ def _transform_cudf_df(
             cat_codes.append(codes)
     else:
         for col in data:
-            if is_categorical_dtype(data[col].dtype) and enable_categorical:
+            dtype = data[col].dtype
+            if is_categorical_dtype(dtype) and enable_categorical:
                 codes = data[col].cat.codes
                 cat_codes.append(codes)
+            elif is_categorical_dtype(dtype):
+                raise ValueError(_ENABLE_CAT_ERR)
+            else:
+                cat_codes.append([])
 
     return data, cat_codes, feature_names, feature_types
 
 
 def _from_cudf_df(
-    data,
-    missing,
-    nthread,
-    feature_names: FeatureNames,
-    feature_types: Optional[List[str]],
+    data: DataType,
+    missing: FloatCompatible,
+    nthread: int,
+    feature_names: Optional[FeatureNames],
+    feature_types: Optional[FeatureTypes],
     enable_categorical: bool,
-) -> Tuple[ctypes.c_void_p, Any, Any]:
+) -> DispatchedDataBackendReturnType:
     data, cat_codes, feature_names, feature_types = _transform_cudf_df(
         data, feature_names, feature_types, enable_categorical
     )
@@ -698,22 +771,18 @@ def _from_cudf_df(
     return handle, feature_names, feature_types
 
 
-def _is_cudf_ser(data):
-    try:
-        import cudf
-    except ImportError:
-        return False
-    return isinstance(data, cudf.Series)
+def _is_cudf_ser(data: DataType) -> bool:
+    return lazy_isinstance(data, "cudf.core.series", "Series")
 
 
-def _is_cupy_array(data) -> bool:
+def _is_cupy_array(data: DataType) -> bool:
     return any(
         lazy_isinstance(data, n, "ndarray")
         for n in ("cupy.core.core", "cupy", "cupy._core.core")
     )
 
 
-def _transform_cupy_array(data):
+def _transform_cupy_array(data: DataType) -> CupyT:
     import cupy  # pylint: disable=import-error
     if not hasattr(data, '__cuda_array_interface__') and hasattr(
             data, '__array__'):
@@ -724,12 +793,12 @@ def _transform_cupy_array(data):
 
 
 def _from_cupy_array(
-    data,
-    missing,
-    nthread,
-    feature_names: FeatureNames,
-    feature_types: Optional[List[str]],
-):
+    data: DataType,
+    missing: FloatCompatible,
+    nthread: int,
+    feature_names: Optional[FeatureNames],
+    feature_types: Optional[FeatureTypes],
+) -> DispatchedDataBackendReturnType:
     """Initialize DMatrix from cupy ndarray."""
     data = _transform_cupy_array(data)
     interface_str = _cuda_array_interface(data)
@@ -743,7 +812,7 @@ def _from_cupy_array(
     return handle, feature_names, feature_types
 
 
-def _is_cupy_csr(data):
+def _is_cupy_csr(data: DataType) -> bool:
     try:
         import cupyx
     except ImportError:
@@ -751,7 +820,7 @@ def _is_cupy_csr(data):
     return isinstance(data, cupyx.scipy.sparse.csr_matrix)
 
 
-def _is_cupy_csc(data):
+def _is_cupy_csc(data: DataType) -> bool:
     try:
         import cupyx
     except ImportError:
@@ -759,11 +828,11 @@ def _is_cupy_csc(data):
     return isinstance(data, cupyx.scipy.sparse.csc_matrix)
 
 
-def _is_dlpack(data):
+def _is_dlpack(data: DataType) -> bool:
     return 'PyCapsule' in str(type(data)) and "dltensor" in str(data)
 
 
-def _transform_dlpack(data):
+def _transform_dlpack(data: DataType) -> bool:
     from cupy import fromDlpack  # pylint: disable=E0401
     assert 'used_dltensor' not in str(data)
     data = fromDlpack(data)
@@ -771,27 +840,27 @@ def _transform_dlpack(data):
 
 
 def _from_dlpack(
-    data,
-    missing,
-    nthread,
-    feature_names: FeatureNames,
-    feature_types: Optional[List[str]],
-):
+    data: DataType,
+    missing: FloatCompatible,
+    nthread: int,
+    feature_names: Optional[FeatureNames],
+    feature_types: Optional[FeatureTypes],
+) -> DispatchedDataBackendReturnType:
     data = _transform_dlpack(data)
     return _from_cupy_array(data, missing, nthread, feature_names,
                             feature_types)
 
 
-def _is_uri(data):
+def _is_uri(data: DataType) -> bool:
     return isinstance(data, (str, os.PathLike))
 
 
 def _from_uri(
-    data,
-    missing,
-    feature_names: FeatureNames,
-    feature_types: Optional[List[str]],
-):
+    data: DataType,
+    missing: Optional[FloatCompatible],
+    feature_names: Optional[FeatureNames],
+    feature_types: Optional[FeatureTypes],
+) -> DispatchedDataBackendReturnType:
     _warn_unused_missing(data, missing)
     handle = ctypes.c_void_p()
     data = os.fspath(os.path.expanduser(data))
@@ -801,51 +870,51 @@ def _from_uri(
     return handle, feature_names, feature_types
 
 
-def _is_list(data):
+def _is_list(data: DataType) -> bool:
     return isinstance(data, list)
 
 
 def _from_list(
-    data,
-    missing,
-    n_threads,
-    feature_names: FeatureNames,
-    feature_types: Optional[List[str]],
-):
+    data: Sequence,
+    missing: FloatCompatible,
+    n_threads: int,
+    feature_names: Optional[FeatureNames],
+    feature_types: Optional[FeatureTypes],
+) -> DispatchedDataBackendReturnType:
     array = np.array(data)
     _check_data_shape(data)
     return _from_numpy_array(array, missing, n_threads, feature_names, feature_types)
 
 
-def _is_tuple(data):
+def _is_tuple(data: DataType) -> bool:
     return isinstance(data, tuple)
 
 
 def _from_tuple(
-    data,
-    missing,
-    n_threads,
-    feature_names: FeatureNames,
-    feature_types: Optional[List[str]],
-):
+    data: Sequence,
+    missing: FloatCompatible,
+    n_threads: int,
+    feature_names: Optional[FeatureNames],
+    feature_types: Optional[FeatureTypes],
+) -> DispatchedDataBackendReturnType:
     return _from_list(data, missing, n_threads, feature_names, feature_types)
 
 
-def _is_iter(data):
+def _is_iter(data: DataType) -> bool:
     return isinstance(data, DataIter)
 
 
-def _has_array_protocol(data):
+def _has_array_protocol(data: DataType) -> bool:
     return hasattr(data, '__array__')
 
 
-def _convert_unknown_data(data):
+def _convert_unknown_data(data: DataType) -> DataType:
     warnings.warn(
         f'Unknown data type: {type(data)}, trying to convert it to csr_matrix',
         UserWarning
     )
     try:
-        import scipy
+        import scipy.sparse
     except ImportError:
         return None
 
@@ -858,13 +927,13 @@ def _convert_unknown_data(data):
 
 
 def dispatch_data_backend(
-    data,
-    missing,
-    threads,
-    feature_names: FeatureNames,
-    feature_types: Optional[List[str]],
+    data: DataType,
+    missing: FloatCompatible,  # Or Optional[Float]
+    threads: int,
+    feature_names: Optional[FeatureNames],
+    feature_types: Optional[FeatureTypes],
     enable_categorical: bool = False,
-):
+) -> DispatchedDataBackendReturnType:
     '''Dispatch data for DMatrix.'''
     if not _is_cudf_ser(data) and not _is_pandas_series(data):
         _check_data_shape(data)
@@ -877,8 +946,7 @@ def dispatch_data_backend(
             data.tocsr(), missing, threads, feature_names, feature_types
         )
     if _is_numpy_array(data):
-        return _from_numpy_array(data, missing, threads, feature_names,
-                                 feature_types)
+        return _from_numpy_array(data, missing, threads, feature_names, feature_types)
     if _is_uri(data):
         return _from_uri(data, missing, feature_names, feature_types)
     if _is_list(data):
@@ -932,7 +1000,7 @@ def dispatch_data_backend(
     raise TypeError('Not supported type for data.' + str(type(data)))
 
 
-def _to_data_type(dtype: str, name: str):
+def _to_data_type(dtype: str, name: str) -> int:
     dtype_map = {'float32': 1, 'float64': 2, 'uint32': 3, 'uint64': 4}
     if dtype not in dtype_map:
         raise TypeError(
@@ -941,7 +1009,7 @@ def _to_data_type(dtype: str, name: str):
     return dtype_map[dtype]
 
 
-def _validate_meta_shape(data: Any, name: str) -> None:
+def _validate_meta_shape(data: DataType, name: str) -> None:
     if hasattr(data, "shape"):
         msg = f"Invalid shape: {data.shape} for {name}"
         if name in _matrix_meta:
@@ -958,7 +1026,7 @@ def _validate_meta_shape(data: Any, name: str) -> None:
 def _meta_from_numpy(
     data: np.ndarray,
     field: str,
-    dtype: Optional[Union[np.dtype, str]],
+    dtype: Optional[NumpyDType],
     handle: ctypes.c_void_p,
 ) -> None:
     data, dtype = _ensure_np_dtype(data, dtype)
@@ -969,16 +1037,26 @@ def _meta_from_numpy(
     _check_call(_LIB.XGDMatrixSetInfoFromInterface(handle, c_str(field), interface_str))
 
 
-def _meta_from_list(data, field, dtype, handle):
-    data = np.array(data)
-    _meta_from_numpy(data, field, dtype, handle)
+def _meta_from_list(
+    data: Sequence,
+    field: str,
+    dtype: Optional[NumpyDType],
+    handle: ctypes.c_void_p
+) -> None:
+    data_np = np.array(data)
+    _meta_from_numpy(data_np, field, dtype, handle)
 
 
-def _meta_from_tuple(data, field, dtype, handle):
+def _meta_from_tuple(
+    data: Sequence,
+    field: str,
+    dtype: Optional[NumpyDType],
+    handle: ctypes.c_void_p
+) -> None:
     return _meta_from_list(data, field, dtype, handle)
 
 
-def _meta_from_cudf_df(data, field: str, handle: ctypes.c_void_p) -> None:
+def _meta_from_cudf_df(data: DataType, field: str, handle: ctypes.c_void_p) -> None:
     if field not in _matrix_meta:
         _meta_from_cudf_series(data.iloc[:, 0], field, handle)
     else:
@@ -987,7 +1065,7 @@ def _meta_from_cudf_df(data, field: str, handle: ctypes.c_void_p) -> None:
         _check_call(_LIB.XGDMatrixSetInfoFromInterface(handle, c_str(field), interface))
 
 
-def _meta_from_cudf_series(data, field, handle):
+def _meta_from_cudf_series(data: DataType, field: str, handle: ctypes.c_void_p) -> None:
     interface = bytes(json.dumps([data.__cuda_array_interface__],
                                  indent=2), 'utf-8')
     _check_call(_LIB.XGDMatrixSetInfoFromInterface(handle,
@@ -995,7 +1073,7 @@ def _meta_from_cudf_series(data, field, handle):
                                                    interface))
 
 
-def _meta_from_cupy_array(data, field, handle):
+def _meta_from_cupy_array(data: DataType, field: str, handle: ctypes.c_void_p) -> None:
     data = _transform_cupy_array(data)
     interface = bytes(json.dumps([data.__cuda_array_interface__],
                                  indent=2), 'utf-8')
@@ -1004,14 +1082,22 @@ def _meta_from_cupy_array(data, field, handle):
                                                    interface))
 
 
-def _meta_from_dt(data, field: str, dtype, handle: ctypes.c_void_p):
+def _meta_from_dt(
+    data: DataType,
+    field: str,
+    dtype: Optional[NumpyDType],
+    handle: ctypes.c_void_p
+) -> None:
     data, _, _ = _transform_dt_df(data, None, None, field, dtype)
     _meta_from_numpy(data, field, dtype, handle)
 
 
 def dispatch_meta_backend(
-    matrix: DMatrix, data, name: str, dtype: Optional[Union[str, np.dtype]] = None
-):
+    matrix: DMatrix,
+    data: DataType,
+    name: str,
+    dtype: Optional[NumpyDType] = None
+) -> None:
     '''Dispatch for meta info.'''
     handle = matrix.handle
     assert handle is not None
@@ -1028,8 +1114,7 @@ def dispatch_meta_backend(
         _meta_from_numpy(data, name, dtype, handle)
         return
     if _is_pandas_df(data):
-        data, _, _ = _transform_pandas_df(data, False, meta=name,
-                                          meta_type=dtype)
+        data, _, _ = _transform_pandas_df(data, False, meta=name, meta_type=dtype)
         _meta_from_numpy(data, name, dtype, handle)
         return
     if _is_pandas_series(data):
@@ -1075,7 +1160,7 @@ class SingleBatchInternalIter(DataIter):  # pylint: disable=R0902
     area for meta info.
 
     '''
-    def __init__(self, **kwargs: Any):
+    def __init__(self, **kwargs: Any) -> None:
         self.kwargs = kwargs
         self.it = 0             # pylint: disable=invalid-name
         super().__init__()
@@ -1092,11 +1177,13 @@ class SingleBatchInternalIter(DataIter):  # pylint: disable=R0902
 
 
 def _proxy_transform(
-    data,
-    feature_names: FeatureNames,
-    feature_types: Optional[List[str]],
+    data: DataType,
+    feature_names: Optional[FeatureNames],
+    feature_types: Optional[FeatureTypes],
     enable_categorical: bool,
-):
+) -> Tuple[
+    Union[bool, ctypes.c_void_p, np.ndarray],
+        Optional[list], Optional[FeatureNames], Optional[FeatureTypes]]:
     if _is_cudf_df(data) or _is_cudf_ser(data):
         return _transform_cudf_df(
             data, feature_names, feature_types, enable_categorical
@@ -1107,6 +1194,7 @@ def _proxy_transform(
     if _is_dlpack(data):
         return _transform_dlpack(data), None, feature_names, feature_types
     if _is_numpy_array(data):
+        data, _ = _ensure_np_dtype(data, data.dtype)
         return data, None, feature_names, feature_types
     if _is_scipy_csr(data):
         return data, None, feature_names, feature_types
@@ -1120,7 +1208,7 @@ def _proxy_transform(
 
 def dispatch_proxy_set_data(
     proxy: _ProxyDMatrix,
-    data: Any,
+    data: DataType,
     cat_codes: Optional[list],
     allow_host: bool,
 ) -> None:
@@ -1130,11 +1218,11 @@ def dispatch_proxy_set_data(
 
     if _is_cudf_df(data):
         # pylint: disable=W0212
-        proxy._set_data_from_cuda_columnar(data, cat_codes)
+        proxy._set_data_from_cuda_columnar(data, cast(List, cat_codes))
         return
     if _is_cudf_ser(data):
         # pylint: disable=W0212
-        proxy._set_data_from_cuda_columnar(data, cat_codes)
+        proxy._set_data_from_cuda_columnar(data, cast(List, cat_codes))
         return
     if _is_cupy_array(data):
         proxy._set_data_from_cuda_interface(data)  # pylint: disable=W0212
